# -*- coding: utf-8 -*-
"""
Created on Wed Aug 21 13:01:11 2019

@author: icruicks
"""

import pandas as pd, GraphLearning as gl, networkx as nx, numpy as np
from scipy.sparse import csr_matrix
from sklearn.metrics import silhouette_score
import os, community
from matplotlib import pyplot as plt


malware = 'sakula'

content = pd.read_csv(os.path.join('Data','Malware', malware+'-dnn2.csv'), index_col=0)

byte_entropies = content.iloc[:,:256]
import_address_table = content.iloc[:,256:512]
string_hashes = content.iloc[:, 512:768]
pe_headers = content.iloc[:,768:1024]

modes = [byte_entropies, import_address_table, string_hashes, pe_headers]

#modal_graphs = [gl.modularity_knn_graph(metric='braycurtis', symmetrize=False).learn_graph(mode) for mode in modes]

#Try applying consensus rule to combined output
#full_graph = modal_graphs[0]
#for graph in modal_graphs[1:]:
#    full_graph = full_graph.maximum(graph)
#final_graph = full_graph
#wcg = gl.weighted_consensus_graph(input_graph = full_graph)
#final_graph = wcg.learn_graph()

#Annealed Graph
#annealed = gl.annealed_graph(scoring='bray_curtis')
#final_graph = annealed.learn_graph(modal_graphs)

#Try cross diffusion on resulting graphs (using modal graph as both kernel and full)
#cross_diffusion = gl.cross_diffusion_graph(kernel_graphs=modal_graphs, T =5, tau=0)
#final_graph = cross_diffusion.learn_graph(modal_graphs)

#Try diffusing on the combined graph of the modal graph
#full_graph = modal_graphs[0]
#sparse_graph = modal_graphs[0]
#for graph in modal_graphs[1:]:
#    full_graph = full_graph.maximum(graph)
#    sparse_graph = sparse_graph.minimum(graph)
#
#diffusion = gl.network_enhancement(tau='mean')
#final_graph = diffusion.learn_graph(full_graph)


# Trying creating a coassociatio matrix for each mode, and then use the kNNs to
# cross-difuse the results to produce a final CO matrix/graph

kNNs =[]
cluster_graphs =[]
clusters = []
ba_matrices =[]

for mode in modes:
    model = gl.modularity_knn_graph(metric='braycurtis', symmetrize=False)
    kNNs.append(model.learn_graph(mode))
    cluster = np.array(list(model.best_part.values()))
    clusters.append(cluster)
    ba_matrix = np.zeros((cluster.size, cluster.max()+1))
    ba_matrix[np.arange(cluster.size),cluster] = 1
    ba_matrices.append(ba_matrix)
    co_matrix = np.matmul(ba_matrix, ba_matrix.T)
    np.fill_diagonal(co_matrix, 0)
    cluster_graphs.append(csr_matrix(co_matrix))

CD = gl.cross_enhancement_graph(kernel_graphs=kNNs, tau=0)
final_graph = CD.learn_graph(cluster_graphs)
    
#final_graph = np.sum(cluster_graphs)/len(cluster_graphs)

#from scipy.spatial.distance import jaccard
#ba_matrix = np.concatenate(ba_matrices, axis=1)
#ca_adjacency = np.zeros((ba_matrix.shape[1], ba_matrix.shape[1]))
#for col_idx in range(ba_matrix.shape[1]):
#    for alt_idx in range(col_idx+1, ba_matrix.shape[1]):
##        ca_adjacency[col_idx, alt_idx] = 1-jaccard(ba_matrix[:,col_idx], ba_matrix[:,alt_idx])
#        ca_adjacency[col_idx, alt_idx] = np.dot(ba_matrix[:,col_idx], ba_matrix[:,alt_idx])/ np.minimum(np.sum(ba_matrix[:,col_idx]), np.sum(ba_matrix[:,alt_idx]))
#        ca_adjacency[alt_idx, col_idx] = ca_adjacency[col_idx, alt_idx]
#    
#ca_graph = nx.from_numpy_array(ca_adjacency)
#meta_clusters = np.array(list(community.best_partition(ca_graph).values()))+1
#unique_meta_clusters = np.unique(meta_clusters)
#aug_matrix = np.multiply(ba_matrix, meta_clusters.T)
#subgroup_counts =[]
#for row in aug_matrix:
#    count_row =[]
#    for cluster in unique_meta_clusters:
#        count_row.append(np.sum(row==cluster))
#    subgroup_counts.append(count_row)
#    
#subgroup_counts = np.array(subgroup_counts)
#final_clusters = subgroup_counts / len(ba_matrices)
#
#final_graph = csr_matrix(np.matmul(final_clusters, final_clusters.T))

G = nx.from_scipy_sparse_matrix(final_graph)
subgroups = community.best_partition(G)
plt.figure()
nx.draw(G, node_color=list(subgroups.values()))

total_set = content.iloc[:,:1024]
print("Silhouette Score: {}".format(silhouette_score(total_set, list(subgroups.values()))))
print("Modularity of the final graph: {}".format(community.modularity(subgroups, G)))














